races = {
    'Zerg': 'Zerg',
    '저그': 'Zerg',
    '异虫': 'Zerg',
    '蟲族': 'Zerg',
    '테란': 'Terran',
    '人類': 'Terran',
    '人类': 'Terran',
    'Terran': 'Terran',
    'Terraner': 'Terran',
    'Терраны': 'Terran',
    '프로토스': 'Protoss',
    '神族': 'Protoss',
    'Protosi': 'Protoss',
    '星灵': 'Protoss',
    'Протоссы': 'Protoss',
    'Protoss': 'Protoss'
}
